/*
    Copyright (c) 2010, 2011 aizo ag, Zurich, Switzerland

    Author: Sergey 'Jin' Bostandzhyan <jin@dev.digitalstrom.org>

    This file is part of dS Assistant.

    dS-Assistant is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    dS-Assistant is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

    NOTE: all APIs, especially the JSON API, offered by this application
          are subject to change without further notice.
*/

#ifndef __DSA_DSM_API_WRAPPER_H__
#define __DSA_DSM_API_WRAPPER_H__

#include <digitalSTROM/dsm-api-v2/dsm-api-const.h>
#include <digitalSTROM/dsm-api-v2/dsm-api.h>

#include <boost/thread/mutex.hpp>
#include <boost/thread/recursive_mutex.hpp>
#include <string>
#include <list>
#include <vector>

#define DSM_API_MAX_DEVICES 63

/// \file dsmapi-wrapper.h
namespace dSA
{

class dSMApiListener
{
public:
    /// dS485-Bus-Level Callbacks
    virtual void busStateCallback(bus_state_t state) = 0;
    virtual void busChangeCallback(dsid_t *id, int flag) = 0;

    /// Device-Level Callbacks
    virtual void deviceAddedCallback(uint8_t error_code, dsid_t src_dsm_dsid,
                                     dsid_t dest_dsm_dsid, uint16_t device_id,
                                     uint16_t zone_id, uint16_t vendor_id,
                                     uint32_t serial_number) = 0;
    virtual void deviceRemovedCallback(uint8_t error_code, dsid_t src_dsm_dsid,
                                       dsid_t dest_dsm_dsid, uint16_t device_id,
                                       uint16_t zone_id, uint16_t vendor_id,
                                       uint32_t serial_number) = 0;
    virtual void deviceStatisticsCallback(uint8_t error_code,
            dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
            uint8_t offset, uint16_t value) = 0;
    virtual void deviceActionCallback(uint8_t error_code,
            dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
            uint16_t zone_id, uint8_t group_id, uint8_t button_nr,
            uint8_t click_type) = 0;
    virtual void deviceLocalActionCallback(uint8_t error_code,
            dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
            uint16_t zone_id, uint8_t state) = 0;
    virtual void deviceConfigCallback(uint8_t error_code,
            dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
            uint8_t config_class, uint8_t config_index, uint8_t value) = 0;
    virtual void deviceTestTransmissionQualityCallback(uint8_t error_code,
            dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
            uint8_t downstream_quality, uint16_t upstream_quality) = 0;

    /// Communication Callbacks
    virtual void commUpstreamShortCallback(uint8_t error_code,
            dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
            uint8_t circuit_id, uint8_t is_resend, uint8_t is_sensor,
            uint8_t key_type, uint8_t click_type,
            uint8_t quality, uint8_t flags, uint8_t cflags) = 0;
    virtual void commUpstreamLongCallback(uint8_t error_code,
            dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
            uint8_t circuit_id, uint8_t mainqualifier, uint8_t subqualifier,
            uint32_t data, uint8_t quality, uint8_t flags, uint8_t cflags) = 0;
    virtual void commDownstreamShortCallback(uint8_t error_code,
            dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint8_t command) = 0;
    virtual void commDownstreamLong0Callback(uint8_t error_code,
            dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint8_t command) = 0;
    virtual void commDownstreamLong8Callback(uint8_t error_code,
            dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint8_t command,
            uint8_t data) = 0;
    virtual void commDownstreamLong16Callback(uint8_t error_code,
            dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint8_t command,
            uint16_t data) = 0;
};

class dSMApiSelectiveListener : public dSMApiListener
{
public:
    typedef enum
    {
        BusState,
        BusChange,
        DeviceAdded,
        DeviceRemoved,
        DeviceStatistics,
        DeviceAction,
        DeviceLocalAction,
        DeviceConfig,
        DeviceTestTransmissionQuality,
        CommUpstreamShort,
        CommUpstreamLong,
        CommDownstreamShort,
        CommDownstreamLong0,
        CommDownstreamLong8,
        CommDownstreamLong16,

        CB_MAX
    } cb_selection;

    class CBData
    {
    public:
        CBData(cb_selection type);
        virtual ~CBData() {}
        cb_selection getType() const;
        time_t getTimestamp() const;
    private:
        CBData();
        cb_selection m_data_type;
        time_t m_timestamp;
    };

    class CBCommonData : public CBData
    {
    public:
        CBCommonData(cb_selection type, uint8_t error_code,
               dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid);
        uint8_t getErrorCode() const;
        dsid_t getDSMSourceDSID() const;
        dsid_t getDSMDestinationDSID() const;
    private:
        uint8_t m_error_code;
        dsid_t m_src_dsm_dsid;
        dsid_t m_dest_dsm_dsid;
    };

    class BasicDeviceData : public CBCommonData
    {
    public:
        BasicDeviceData(cb_selection type, uint8_t error_code,
                dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id);
        uint16_t getDeviceID() const;
    private:
        uint16_t m_device_id;
    };

    class DeviceData : public BasicDeviceData
    {
    public:
        DeviceData(cb_selection type, uint8_t error_code,
                dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
                uint16_t zone_id, uint16_t vendor_id,
                uint32_t serial_number);
        uint16_t getZoneID() const;
        uint16_t getVendorID() const;
        uint16_t getSerialNumber() const;
    private:
        uint16_t m_zone_id;
        uint16_t m_vendor_id;
        uint32_t m_serial_number;
    };

    class DeviceAddedData: public DeviceData
    {
    public:
        DeviceAddedData(uint8_t error_code, dsid_t src_dsm_dsid,
                dsid_t dest_dsm_dsid, uint16_t device_id, uint16_t zone_id,
                uint16_t vendor_id, uint32_t serial_number);
    };

    class DeviceRemovedData: public DeviceData
    {
    public:
        DeviceRemovedData(uint8_t error_code, dsid_t src_dsm_dsid,
                dsid_t dest_dsm_dsid, uint16_t device_id, uint16_t zone_id,
                uint16_t vendor_id, uint32_t serial_number);
    };

    class DeviceTestQualityData: public BasicDeviceData
    {
    public:
        DeviceTestQualityData(uint8_t error_code, dsid_t src_dsm_dsid,
                dsid_t dest_dsm_dsid, uint16_t device_id, uint8_t down, uint16_t up);
        uint8_t getDownstream() const;
        uint16_t getUpstream() const;
    private:
        uint8_t m_down;
        uint16_t m_up;
    };

    class DeviceStatisticData: public BasicDeviceData
    {
    public:
        DeviceStatisticData(uint8_t error_code, dsid_t src_dsm_dsid,
                dsid_t dest_dsm_dsid, uint16_t device_id,
                uint8_t offset, uint16_t value);
        uint8_t getOffset() const;
        uint16_t getValue() const;
    private:
        uint8_t m_offset;
        uint16_t m_value;
    };

    class CommData : public BasicDeviceData
    {
    public:
        CommData(cb_selection type, uint8_t error_code, dsid_t src_dsm_dsid,
                 dsid_t dest_dsm_dsid, uint16_t device_id,
                 uint8_t circuit_id, uint8_t quality,
                 uint8_t flags, uint8_t cflags);
        uint8_t getCircuitID() const;
        uint8_t getQuality() const;
        uint8_t getFlags() const;
        uint8_t getCrosstalkFlags() const;
    private:
        uint8_t m_circuit_id;
        uint8_t m_quality;
        uint8_t m_flags;
        uint8_t m_cflags;
    };

    class CommUpstreamShortData : public CommData
    {
    public:
        CommUpstreamShortData(uint8_t error_code, dsid_t src_dsm_dsid,
            dsid_t dest_dsm_dsid, uint16_t device_id,
            uint8_t circuit_id, uint8_t is_resend, uint8_t is_sensor,
            uint8_t key_type, uint8_t click_type, uint8_t quality,
            uint8_t flags, uint8_t cflags);
        uint8_t isResend() const;
        uint8_t isSensor() const;
        uint8_t getKeyType() const;
        uint8_t getClickType() const;

    private:
        uint8_t m_is_resend;
        uint8_t m_is_sensor;
        uint8_t m_key_type;
        uint8_t m_click_type;
    };

    class CommUpstreamLongData : public CommData
    {
    public:
        CommUpstreamLongData(uint8_t error_code,
            dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
            uint8_t circuit_id, uint8_t mainqualifier, uint8_t subqualifier,
            uint32_t data, uint8_t quality, uint8_t flags, uint8_t cflags);
        uint8_t getMainQualifier() const;
        uint8_t getSubQualifier() const;
        uint32_t getData() const;

    private:
        uint8_t m_mainqualifier;
        uint8_t m_subqualifier;
        uint32_t m_data;
    };

    class CommDownstreamData : public CBCommonData
    {
    public:
        CommDownstreamData(cb_selection type, uint8_t error_code,
                dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint8_t command);
        uint8_t getCommand() const;
    private:
        uint8_t m_command;
    };

    class CommDownstreamShortData : public CommDownstreamData
    {
    public:
        CommDownstreamShortData(uint8_t error_code, dsid_t src_dsm_dsid,
                dsid_t dest_dsm_dsid, uint8_t command);
    };

    class CommDownstreamLong0Data : public CommDownstreamData
    {
    public:
        CommDownstreamLong0Data(uint8_t error_code, dsid_t src_dsm_dsid,
                dsid_t dest_dsm_dsid, uint8_t command);
    };

    class CommDownstreamLong8Data : public CommDownstreamData
    {
    public:
        CommDownstreamLong8Data(uint8_t error_code, dsid_t src_dsm_dsid,
                dsid_t dest_dsm_dsid, uint8_t command, uint8_t data);
        uint8_t getData() const;
    private:
        uint8_t m_data;
    };

    class CommDownstreamLong16Data : public CommDownstreamData
    {
    public:
        CommDownstreamLong16Data(uint8_t error_code, dsid_t src_dsm_dsid,
                dsid_t dest_dsm_dsid, uint8_t command, uint16_t data);
        uint16_t getData() const;
    private:
        uint16_t m_data;
    };

    dSMApiSelectiveListener();
    virtual ~dSMApiSelectiveListener();
    void select(dSMApiSelectiveListener::cb_selection cb, bool on);

    virtual void busStateCallback(bus_state_t state);
    virtual void busChangeCallback(dsid_t *id, int flag);

    /// Device-Level Callbacks
    virtual void deviceAddedCallback(uint8_t error_code, dsid_t src_dsm_dsid,
                                     dsid_t dest_dsm_dsid, uint16_t device_id,
                                     uint16_t zone_id, uint16_t vendor_id,
                                     uint32_t serial_number);
    virtual void deviceRemovedCallback(uint8_t error_code, dsid_t src_dsm_dsid,
                                       dsid_t dest_dsm_dsid, uint16_t device_id,
                                       uint16_t zone_id, uint16_t vendor_id,
                                       uint32_t serial_number);
    virtual void deviceStatisticsCallback(uint8_t error_code,
            dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
            uint8_t offset, uint16_t value);
    virtual void deviceActionCallback(uint8_t error_code,
            dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
            uint16_t zone_id, uint8_t group_id, uint8_t button_nr,
            uint8_t click_type);
    virtual void deviceLocalActionCallback(uint8_t error_code,
            dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
            uint16_t zone_id, uint8_t state);
    virtual void deviceConfigCallback(uint8_t error_code,
            dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
            uint8_t config_class, uint8_t config_index, uint8_t value);
    virtual void deviceTestTransmissionQualityCallback(uint8_t error_code,
            dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
            uint8_t downstream_quality, uint16_t upstream_quality);

    /// Communication Callbacks
    virtual void commUpstreamShortCallback(uint8_t error_code,
            dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
            uint8_t circuit_id, uint8_t is_resend, uint8_t is_sensor,
            uint8_t key_type, uint8_t click_type, uint8_t quality,
            uint8_t flags, uint8_t cflags);
    virtual void commUpstreamLongCallback(uint8_t error_code,
            dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
            uint8_t circuit_id, uint8_t mainqualifier, uint8_t subqualifier,
            uint32_t data, uint8_t quality, uint8_t flags, uint8_t cflags);
    virtual void commDownstreamShortCallback(uint8_t error_code,
            dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint8_t command);
    virtual void commDownstreamLong0Callback(uint8_t error_code,
            dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint8_t command);
    virtual void commDownstreamLong8Callback(uint8_t error_code,
            dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint8_t command,
            uint8_t data);
    virtual void commDownstreamLong16Callback(uint8_t error_code,
            dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint8_t command,
            uint16_t data);

    virtual void callback(boost::shared_ptr<CBData> callback_data) = 0;
private:
    bool checkCB(cb_selection sel);
    std::vector<bool> m_selection;
    boost::recursive_mutex m_sel_access_mutex;
};

/// This class provides a threadsafe access to the dSM API library.
///
/// Each new instance of the class will open a new connection to the dS485
/// daemon. However each instance is threadsafe and can be used from various
/// threads. Currently it will only wrap the functions that we are interested
/// in, more calls will be added on demand.
///
/// Instead of return values the wrapped calls to the lib will throw exceptions.
class dSMApiWrapper
{
public:
    dSMApiWrapper(std::string ds485_daemon_url);
    ~dSMApiWrapper();

    void addListener(dSMApiListener* listener);
    void removeListener(dSMApiListener* listener);

    void getOwnDSID(dsid_t* dsid);
    bus_state_t getBusState();
    int getBusMembers(dsid_t *device_list, int device_list_size);

    /// dSMeter Infos

    void dSMInfo(dsid_t dsm_dsid, uint32_t *hw_version,
                 uint32_t *arm_sw_version, uint32_t *dsp_sw_version,
                 uint16_t *api_version, dsid_t *dsid, uint8_t *name);
    void dSMState(dsid_t dsm_dsid, uint8_t *hardware, uint8_t *communicationdsp,
                  uint8_t *temperature, uint8_t *configuration,
                  uint8_t *ds485bus, uint8_t *powerline);
    void getPropertyFlags(dsid_t dsm_dsid, uint8_t *flags);
    void setPropertyFlags(dsid_t dsm_dsid, uint8_t flags);
    void getPropertyTemperatureLimit(dsid_t dsm_dsid, uint8_t *warnlevel, uint8_t *errorlevel);
    void setPropertyTemperatureLimit(dsid_t dsm_dsid, uint8_t warnlevel, uint8_t errorlevel);
    void getDSMtemperature(dsid_t dsm_dsid, uint16_t *temperature);
    void getApartmentId(dsid_t dsm_dsid, uint8_t *id);
    void setApartmentId(dsid_t dsm_dsid, uint8_t id);
    void getConfigurationId(dsid_t dsm_dsid, uint32_t *hash,
                            uint32_t *configurationcount);
    void getStatisticUpstream(dsid_t dsm_dsid, uint8_t index,
            uint8_t *length, uint8_t *statistics);
    void resetStatisticUpstream(dsid_t dsm_dsid);

    /// Infrastructure

    void getZoneCount(dsid_t dsm_dsid, uint8_t *count);
    void getZoneInfo(dsid_t dsm_dsid, uint8_t zone_index, uint16_t *zone_id,
                      uint8_t *virtual_zone_id, uint8_t *number_of_groups,
                      uint8_t *name);
    void getActiveDevicesCountInZone(dsid_t dsm_dsid, uint16_t zone_id,
                                     uint16_t *number_of_devices);
    void getInactiveDevicesCountInZone(dsid_t dsm_dsid, uint16_t zone_id,
                                       uint16_t *number_of_devices);
    void getDevicesCountInZone(dsid_t dsm_dsid, uint16_t zone_id,
                               uint16_t *number_of_devices);
    void getActiveDeviceInfoInZone(dsid_t dsm_dsid, uint16_t req_zone_id,
                             uint16_t device_index, uint16_t *device_id,
                             uint16_t *vendor_id, uint16_t *product_id,
                             uint16_t *function_id, uint16_t *version,
                             uint16_t *zone_id, uint8_t *state,
                             uint8_t *my_click, uint8_t *locked,
                             uint8_t *output_mode, uint8_t *groups,
                             uint8_t *name, uint8_t *buttonGroups,
                             uint8_t *button_zone_membership,
                             uint32_t *serial_number);
    void getDeviceInfoById(dsid_t dsm_dsid, uint16_t req_device_id,
                       uint16_t *device_id, uint16_t *vendor_id,
                       uint16_t *product_id, uint16_t *function_id,
                       uint16_t *version, uint16_t *zone_id, uint8_t *state,
                       uint8_t *button_id, uint8_t *locked,
                       uint8_t *output_mode, uint8_t *groups,
                       uint8_t *name, uint8_t *button_groups,
                       uint8_t *button_zone_membership,
                       uint32_t *serial_number);
    void getDeviceInfoByIndex(dsid_t dsm_dsid, uint16_t req_zone_id,
                       uint16_t req_index,
                       uint16_t *device_id, uint16_t *vendor_id,
                       uint16_t *product_id, uint16_t *function_id,
                       uint16_t *version, uint16_t *zone_id, uint8_t *state,
                       uint8_t *button_id, uint8_t *locked,
                       uint8_t *output_mode, uint8_t *groups,
                       uint8_t *name, uint8_t *button_groups,
                       uint8_t *button_zone_membership,
                       uint32_t *serial_number);

    // dSMeter Actions
    void restartRegistration(dsid_t dsm_dsid);
    void restartMeter(dsid_t dsm_dsid);
    void setDisplayMode(dsid_t dsm_dsid, uint8_t flags);
    void getDisplayMode(dsid_t dsm_dsid, uint8_t* flags);
    void setCommunicationLogging(dsid_t dsm_dsid, uint8_t state);
    void getCommunicationLogging(dsid_t dsm_dsid, uint8_t* state);

    // dSMeter device presence check
    void pingAutomatic(dsid_t dsm_dsid, uint16_t auto_interval_time);
    void pingOneTimeAutomatic(dsid_t dsm_dsid, uint16_t interval_time,
                              uint16_t auto_interval_time);
    void pingOneTimeOff(dsid_t dsm_dsid, uint16_t interval_time);
    void pingOff(dsid_t dsm_dsid);
    void getPingIntervals(dsid_t dsm_dsid, uint16_t *interval_auto,
                          uint16_t *interval_device);
    void setPingIntervals(dsid_t dsm_dsid, uint16_t interval_auto,
                          uint16_t interval_device);

    /// Device Actions
    void blinkDevice(dsid_t dsm_dsid, uint16_t device_id);
    void getConfigurationValue(dsid_t dsm_dsid, uint16_t device_id,
            uint8_t config_class, uint8_t config_index, uint8_t *value);
    void getConfigurationValue(dsid_t dsm_dsid, uint16_t device_id,
            uint8_t config_class, uint8_t config_index, uint16_t *value);
    void getStatisticValue(dsid_t dsm_dsid, uint16_t device_id,
            uint8_t offset, uint8_t *value);
    void getStatisticValue(dsid_t dsm_dsid, uint16_t device_id,
            uint8_t offset, uint16_t *value);
    void getStatisticValue(dsid_t dsm_dsid, uint16_t device_id,
            uint8_t offset, uint32_t *value);
    void reqStatisticValue(dsid_t dsm_dsid, uint16_t device_id,
            uint8_t offset);
    void getSensorValue(dsid_t dsm_dsid, uint16_t device_id,
            uint8_t index, uint16_t *value);
    void getSensorType(dsid_t dsm_dsid, uint16_t device_id,
                       uint8_t index, uint8_t *sensor_present,
                       uint8_t *sensor_type, uint8_t sensor_last);
    void testDeviceTransmissionQuality(dsid_t dsm_dsid, uint16_t device_id,
                                        uint8_t *downstreamQ,
                                        uint16_t *upstreamQ);
    void reqDeviceTransmissionQuality(dsid_t dsm_dsid, uint16_t device_id);

    static std::string error2Message(int err);

private:
    boost::mutex m_libdsm_mutex;
    boost::mutex m_listener_mutex;
    std::string m_daemon_url;
    DsmApiHandle_t m_dsm_api_handle;
    std::list<dSMApiListener*> m_callback_listeners;

    void checkRetcode(int ret);

    static void busStateCallback(void *arg, bus_state_t state);
    void notifyBusState(bus_state_t state);

    static void busChangeCallback(void *arg, dsid_t *id, int flag);
    void notifyBusChange(dsid_t *id, int flag);

    static void deviceAddedCallback(uint8_t error_code, void *arg,
            dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid,
            uint16_t device_id, uint16_t zone_id,
            uint16_t vendor_id, uint32_t serial_number);
    void notifyDeviceAdded(uint8_t error_code,
            dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
            uint16_t zone_id, uint16_t vendor_id, uint32_t serial_number);

    static void deviceRemovedCallback(uint8_t error_code, void *arg,
            dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
            uint16_t zone_id, uint16_t vendor_id, uint32_t serial_number);
    void notifyDeviceRemoved(uint8_t error_code,
            dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
            uint16_t zone_id, uint16_t vendor_id, uint32_t serial_number);

    static void deviceStatisticsCallback(uint8_t error_code, void *arg,
            dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
            uint8_t offset, uint16_t value);
    void notifyDeviceStatistics(uint8_t error_code,
            dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
            uint8_t offset, uint16_t value);

    static void deviceActionCallback(uint8_t error_code, void *arg,
            dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
            uint16_t zone_id, uint8_t group_id, uint8_t button_nr,
            uint8_t click_type);
    void notifyDeviceActionCallback(uint8_t error_code,
            dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
            uint16_t zone_id, uint8_t group_id, uint8_t button_nr,
            uint8_t click_type);

    static void deviceLocalActionCallback(uint8_t error_code, void *arg,
            dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
            uint16_t zone_id, uint8_t state);
    void notifyDeviceLocalActionCallback(uint8_t error_code,
            dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
            uint16_t zone_id, uint8_t state);

    static void deviceConfigCallback(uint8_t error_code, void *arg,
            dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
            uint8_t config_class, uint8_t config_index, uint8_t value);
    void notifyDeviceConfigCallback(uint8_t error_code,
            dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
            uint8_t config_class, uint8_t config_index, uint8_t value);

    static void deviceTestTransmissionQualityCallback(uint8_t error_code,
            void *arg, dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid,
            uint16_t device_id, uint8_t downstream_quality,
            uint16_t upstream_quality);
    void notifyDeviceTestTransmissionQualityCallback(uint8_t error_code,
            dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
            uint8_t downstream_quality, uint16_t upstream_quality);

    static void commUpstreamShortCallback(uint8_t error_code, void *arg,
            dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
            uint8_t circuit_id, uint8_t is_resend, uint8_t is_sensor,
            uint8_t key_type, uint8_t click_type,
            uint8_t quality, uint8_t flags, uint8_t cflags);
    void notifyCommUpstreamShortCallback(uint8_t error_code,
            dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
            uint8_t circuit_id, uint8_t is_resend, uint8_t is_sensor,
            uint8_t key_type, uint8_t click_type,
            uint8_t quality, uint8_t flags, uint8_t cflags);

    static void commUpstreamLongCallback(uint8_t error_code, void *arg,
            dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
            uint8_t circuit_id, uint8_t mainqualifier, uint8_t subqualifier,
            uint32_t data, uint8_t quality, uint8_t flags, uint8_t cflags);
    void notifyCommUpstreamLongCallback(uint8_t error_code,
            dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint16_t device_id,
            uint8_t circuit_id, uint8_t mainqualifier, uint8_t subqualifier,
            uint32_t data, uint8_t quality, uint8_t flags, uint8_t cflags);

    static void commDownstreamShortCallback(uint8_t error_code, void *arg,
            dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint8_t command);
    void notifyCommDownstreamShortCallback(uint8_t error_code,
            dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint8_t command);

    static void commDownstreamLong0Callback(uint8_t error_code, void *arg,
            dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint8_t command);
    void notifyCommDownstreamLong0Callback(uint8_t error_code,
            dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint8_t command);

    static void commDownstreamLong8Callback(uint8_t error_code, void *arg,
            dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint8_t command,
            uint8_t data);
    void notifyCommDownstreamLong8Callback(uint8_t error_code,
            dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint8_t command,
            uint8_t data);

    static void commDownstreamLong16Callback(uint8_t error_code, void *arg,
            dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint8_t command,
            uint16_t data);
    void notifyCommDownstreamLong16Callback(uint8_t error_code,
            dsid_t src_dsm_dsid, dsid_t dest_dsm_dsid, uint8_t command,
            uint16_t data);
};

}; // namespace

#endif//__DSA_DSM_API_WRAPPER_H__

